from io import BytesIO
from pathlib import Path

from astropy.io import fits

from dkist_header_validator import spec214_l0_validator
from dkist_header_validator import spec214_validator


def test_spec214l0(valid_spec_214l0_object):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214l0 fits header
    When: Validating headers
    Then: return validated HDUList and do not raise an exception
    """
    # raises exception on failure
    spec214_l0_validator.validate(valid_spec_214l0_object, extra=False)


def test_spec214(valid_spec_214_object):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: return validated HDUList and do not raise an exception
    """
    # raises exception on failure
    spec214_validator.validate(valid_spec_214_object, extra=False)


def test_spec214l0_return_dictionary(valid_spec_214l0_object):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: return validated dictionary and do not raise an exception
    """
    # raises exception on failure
    result = spec214_l0_validator.validate(valid_spec_214l0_object, return_type=dict, extra=False)
    assert isinstance(result, dict)


def test_spec214_return_dictionary(valid_spec_214_object):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: return validated dictionary and do not raise an exception
    """
    # raises exception on failure
    result = spec214_validator.validate(valid_spec_214_object, return_type=dict, extra=False)
    assert isinstance(result, dict)


def test_spec214l0_return_fits_header(valid_spec_214l0_object):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: return validated fits.header.Header object and do not raise an exception
    """
    # raises exception on failure
    result = spec214_l0_validator.validate(
        valid_spec_214l0_object, return_type=fits.header.Header, extra=False
    )
    assert isinstance(result, fits.header.Header)


def test_spec214_return_fits_header(valid_spec_214_object):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: return validated fits.header.Header object and do not raise an exception
    """
    # raises exception on failure
    result = spec214_validator.validate(
        valid_spec_214_object, return_type=fits.header.Header, extra=False
    )
    assert isinstance(result, fits.header.Header)


def test_spec214l0_return_HDU(valid_spec_214l0_file):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: return validated fits.PrimaryHDU object and do not raise an exception
    """
    # raises exception on failure
    result = spec214_l0_validator.validate(
        valid_spec_214l0_file, return_type=fits.PrimaryHDU, extra=False
    )
    assert isinstance(result, fits.PrimaryHDU)


def test_spec214_return_HDU(valid_spec_214_file):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: return validated fits.PrimaryHDU object and do not raise an exception
    """
    # raises exception on failure
    result = spec214_validator.validate(
        valid_spec_214_file, return_type=fits.PrimaryHDU, extra=False
    )
    assert isinstance(result, fits.PrimaryHDU)


def test_spec214l0_return_BytesIO(valid_spec_214l0_file):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: return validated BytesIO object and do not raise an exception
    """
    # raises exception on failure
    result = spec214_l0_validator.validate(valid_spec_214l0_file, return_type=BytesIO, extra=False)
    assert isinstance(result, BytesIO)


def test_spec214_return_BytesIO(valid_spec_214_file):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: return validated BytesIO object and do not raise an exception
    """
    # raises exception on failure
    result = spec214_validator.validate(valid_spec_214_file, return_type=BytesIO, extra=False)
    assert isinstance(result, BytesIO)


def test_spec214l0_return_file(valid_spec_214l0_file):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: return validated file and do not raise an exception
    """
    # raises exception on failure
    result = spec214_l0_validator.validate(valid_spec_214l0_file, return_type=Path, extra=False)
    assert isinstance(result, Path)


def test_spec214_return_file(valid_spec_214_file):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: return validated file and do not raise an exception
    """
    # raises exception on failure
    result = spec214_validator.validate(valid_spec_214_file, return_type=Path, extra=False)
    assert isinstance(result, Path)


def test_spec214l0_extraheaders_allowed(valid_spec_214l0_object_extra_keys):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: return validated HDUList and do not raise an exception
    """
    # raises exception on failure
    spec214_l0_validator.validate(valid_spec_214l0_object_extra_keys)


def test_spec214_extraheaders_allowed(valid_spec_214_object_extra_keys):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: return validated HDUList and do not raise an exception
    """
    # raises exception on failure
    spec214_validator.validate(valid_spec_214_object_extra_keys)


def test_validatel0_required_only_headers(valid_spec_214l0_object_required_only):
    """
    Validates a spec214l0 compliant header with only the keywords required by the DC
    Given: A spec214l0 compliant header with only required keywords
    When: Validating headers
    Then: return a validated HDUList and do not raise an exception
    """
    spec214_l0_validator.validate(valid_spec_214l0_object_required_only)


def test_validate_required_only_headers(valid_spec_214_object_required_only):
    """
    Validates a spec214 compliant header with only the keywords required by the DC
    Given: A spec214 compliant header with only required keywords
    When: Validating headers
    Then: return a validated HDUList and do not raise an exception
    """
    spec214_validator.validate(valid_spec_214_object_required_only)


def test_validatel0_expected_only_headers(valid_spec_214l0_object_expected_only):
    """
    Validates a spec214l0 compliant header with only the keywords required by the DC
    Given: A spec214l0 compliant header with only required keywords
    When: Validating headers
    Then: return a validated HDUList and do not raise an exception
    """
    spec214_l0_validator.validate(valid_spec_214l0_object_expected_only)


def test_validate_expected_only_headers(valid_spec_214_object_expected_only):
    """
    Validates a spec214 compliant header with only the keywords required by the DC
    Given: A spec214 compliant header with only required keywords
    When: Validating headers
    Then: return a validated HDUList and do not raise an exception
    """
    spec214_validator.validate(valid_spec_214_object_expected_only)


def test_compressed_spec214l0(valid_spec_214l0_compressed):
    """
    Validates a compressed spec214 compliant file
    Given: A valid compressed SPEC-0214 file
    When: Validating headers
    Then: return validated HDUList and do not raise an exception
    """
    spec214_l0_validator.validate(valid_spec_214l0_compressed)


def test_compressed_spec214(valid_spec_214_compressed):
    """
    Validates a compressed spec214 compliant file
    Given: A valid compressed SPEC-0214 file
    When: Validating headers
    Then: return validated HDUList and do not raise an exception
    """
    spec214_validator.validate(valid_spec_214_compressed)


def test_validatel0_datainsecondHDU(valid_spec_214l0_second_hdu):
    """
    Validates headers with data stored in second HDU
    Given: A valid SPEC-214 file or with data stored in second HDU
    When: Validating headers
    Then: Raise an exception
    """
    # raises exception on failure
    spec214_l0_validator.validate(valid_spec_214l0_second_hdu, return_type=Path)


def test_validate_datainsecondHDU(valid_spec_214_second_hdu):
    """
    Validates headers with data stored in second HDU
    Given: A valid SPEC-214 file or with data stored in second HDU
    When: Validating headers
    Then: Do not raise an exception
    """
    # raises exception on failure
    spec214_validator.validate(valid_spec_214_second_hdu, return_type=Path)


def test_validate_s214l0_visp(valid_spec_214l0_visp):
    """
    Validates headers with visp headers and data
    Given: A valid SPEC-214 visp file
    When: Validating headers
    Then: Do not raise an exception
    """
    # raises exception on failure
    spec214_l0_validator.validate(valid_spec_214l0_visp)


def test_validate_s214_visp(valid_spec_214_visp):
    """
    Validates headers with visp headers and data
    Given: A valid SPEC-214 visp file
    When: Validating headers
    Then: Do not raise an exception
    """
    # raises exception on failure
    spec214_validator.validate(valid_spec_214_visp)


def test_validate_s214l0_no_file(valid_spec_214l0_no_file):
    """
    Validates headers of the type hdulist, dict, or fits.Header
    Given: A valid SPEC-214 header
    When: Validating headers
    Then: Do not raise an exception
    """
    # raises exception on failure
    spec214_l0_validator.validate(valid_spec_214l0_no_file)


def test_validate_s214_no_file(valid_spec_214_no_file):
    """
    Validates headers of the type hdulist, dict, or fits.Header
    Given: A valid SPEC-214 header
    When: Validating headers
    Then: Do not raise an exception
    """
    # raises exception on failure
    spec214_validator.validate(valid_spec_214_no_file)


def test_validate_s214l0_dict_only(valid_spec_214l0_dict_only):
    """
    Validates a dict of input S214 headers
    Given: A valid SPEC-214 header
    When: Validating headers
    Then: Do not raise an exception
    """
    # raises exception on failure
    spec214_l0_validator.validate(valid_spec_214l0_dict_only)


def test_validate_s214_dict_only(valid_spec_214_dict_only):
    """
    Validates a dict of input S214 headers
    Given: A valid SPEC-214 header
    When: Validating headers
    Then: Do not raise an exception
    """
    # raises exception on failure
    spec214_validator.validate(valid_spec_214_dict_only)


def test_validate_s214l0_hdulist_only(valid_spec_214l0_hdulist_only):
    """
    Validates an HDUList of input S214 headers
    Given: A valid SPEC-214 header
    When: Validating headers
    Then: Do not raise an exception
    """
    # raises exception on failure
    spec214_l0_validator.validate(valid_spec_214l0_hdulist_only)


def test_validate_s214_hdulist_only(valid_spec_214_hdulist_only):
    """
    Validates an HDUList of input S214 headers
    Given: A valid SPEC-214 header
    When: Validating headers
    Then: Do not raise an exception
    """
    # raises exception on failure
    spec214_validator.validate(valid_spec_214_hdulist_only)


def test_validate_s214l0_casesensitive(valid_spec_214l0_casesensitive):
    """
    Validates headers with a keyword that will need a case sensitive change
    Given: S214 headers with a keyword that will need a case sensitive change
    When: Validating headers
    Then: Do not raise an exception
    """
    # raises exception on failure
    spec214_l0_validator.validate(valid_spec_214l0_casesensitive)


def test_validate_s214_casesensitive(valid_spec_214_casesensitive):
    """
    Validates headers with a keyword that will need a case sensitive change
    Given: S214 headers with a keyword that will need a case sensitive change
    When: Validating headers
    Then: Do not raise an exception
    """
    # raises exception on failure
    spec214_validator.validate(valid_spec_214_casesensitive)



import typer

from dkist_header_validator.spec_validators import spec122_validator
from dkist_header_validator.spec_validators import Spec122ValidationException
from dkist_header_validator.spec_validators import spec214_validator
from dkist_header_validator.spec_validators import Spec214ValidationException

app = typer.Typer(help="CLI for the dkist-header-validator package")


@app.command("validate122")
def validate122(
    filename: str = typer.Argument(help="Location of the FITS file on disk", default=None)
):
    """
    Validate a FITS file for SPEC-0122 compliance.
    SPEC failures are:
    - Extra keywords not included in SPEC-0122
    - Keywords with values of the wrong type (float instead of string, for example)
    - SPEC-0122 keywords that are missing
    - SPEC-0122 keywords with invalid values compared to the valid list given in the SPEC
    """
    try:
        spec122_validator.validate(filename, extra=False)
        print(f"No validation errors found in file {filename}")
    except Spec122ValidationException as e:
        print(f"Validation errors in file {filename}")
        for key, value in e.errors.items():
            print(f"{key}: {value}")


@app.command("validate214")
def validate214(
    filename: str = typer.Argument(help="Location of the FITS file on disk", default=None)
):
    """
    Validate a FITS file for SPEC-0214 compliance.
    SPEC failures are:
    - Extra keywords not included in SPEC-0214
    - Keywords with values of the wrong type (float instead of string, for example)
    - SPEC-0214 keywords that are missing
    - SPEC-0214 keywords with invalid values compared to the valid list given in the SPEC"""
    try:
        spec214_validator.validate(filename, extra=False)
        print(f"No validation errors found in file {filename}")
    except Spec214ValidationException as e:
        print(f"Validation errors in file {filename}")
        for key, value in e.errors.items():
            print(f"{key}: {value}")


def main():
    app()


if __name__ == "__main__":
    main()



import logging
from datetime import datetime
from functools import reduce
from typing import Any
from typing import Dict
from typing import IO
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from astropy.io import fits
from astropy.io.fits.hdu.hdulist import HDUList
from dkist_fits_specifications.spec214 import expand_index_d
from dkist_fits_specifications.spec214 import level0
from dkist_fits_specifications.spec214 import load_expanded_spec214
from dkist_fits_specifications.utils import expand_naxis


logger = logging.getLogger(__name__)

__all__ = [
    "translate_spec122_to_spec214_l0",
    "sanitize_to_spec214_level1",
]

type_map = {"int": int, "float": float, "str": str, "bool": bool}


def translate_spec122_to_spec214_l0(
    spec122_input: Union[HDUList, dict, fits.header.Header, str, IO, List]
) -> Union[dict, HDUList]:
    """
    Convert spec 122 headers to spec 214 l0 headers
    :param spec122_input: Spec 122 headers or headers + data to convert
    """
    # extract headers and data
    input_headers, input_data = _parse_fits_like_input(spec122_input)
    # convert headers
    output_headers = _add_214_l0_headers(input_headers)
    # update DATE keyword
    output_headers["DATE"] = datetime.now().isoformat()
    if input_data:  # return hdu list if the input had data
        return _format_output_hdu(output_headers, input_data)
    return output_headers  # return headers if only headers were given


def sanitize_to_spec214_level1(
    input_headers: Union[HDUList, dict, fits.header.Header, str, IO, List]
) -> Union[dict, HDUList]:
    # extract headers and data
    input_headers, input_data = _parse_fits_like_input(input_headers)
    header = fits.Header(input_headers)
    # convert headers
    expanded_214 = load_expanded_spec214(**dict(input_headers))
    all_214_keys = reduce(list.__add__, map(list, expanded_214.values()))

    for keyword in tuple(header.keys()):
        if keyword not in all_214_keys:
            header.remove(keyword)

    if input_data is not None:  # return hdu list if the input had data
        return _format_output_hdu(header, input_data)
    return header  # return headers if only headers were given


def _parse_fits_like_input(
    spec122_input: Union[HDUList, dict, fits.header.Header, str, IO, List]
) -> Tuple[fits.Header, Optional[bytes]]:
    """
    Parse out a header and optional data from the various types of input
    """
    if isinstance(spec122_input, dict):
        return fits.Header(spec122_input), None
    if isinstance(spec122_input, fits.header.Header):
        return spec122_input, None
    if isinstance(spec122_input, HDUList):
        try:
            return spec122_input[1].header, spec122_input[1].data
        except IndexError:  # non-compressed
            return spec122_input[0].header, spec122_input[0].data

    # If headers are of any other type, see if it is a file and try to open that
    try:  # compressed
        with fits.open(spec122_input) as hdus:
            return hdus[1].header, hdus[1].data
    except IndexError:  # non-compressed
        with fits.open(spec122_input) as hdus:
            return hdus[0].header, hdus[0].data


def _format_output_hdu(hdr, data) -> HDUList:
    new_hdu = fits.PrimaryHDU(data)
    hdu_list = fits.HDUList([new_hdu])
    for key, value in hdr.items():
        hdu_list[0].header[key] = value
    return hdu_list


def _create_spec214_schema_for_header(
    hdr: dict, generic_schema, naxis=None
) -> Dict[str, Dict[str, Any]]:
    s = {key: schema for definition in generic_schema for (key, schema) in definition.items()}
    if naxis == None:
        return s
    else:
        hdr["DNAXIS"] = naxis
        return expand_index_d(expand_naxis(hdr["NAXIS"], s), DNAXIS=hdr["DNAXIS"])


def _add_214_l0_headers(hdr: dict):
    """
    Translates 122 keywords to 214 l0 keywords and returns a dictionary
    """
    result = {}  # output headers
    generic_spec214_l0_schema = level0.load_level0_spec214().values()
    spec214_l0_schema = _create_spec214_schema_for_header(hdr, generic_spec214_l0_schema)

    # translate 122 -> 214 headers
    for key, key_schema in spec214_l0_schema.items():
        result.update(_translate_key(key, key_schema, hdr))

    # add remaining header values to result
    hdr_keys_not_translated = {k: v for k, v in hdr.items() if k not in result}
    result.update(hdr_keys_not_translated)
    return result


def _translate_key(key, key_schema, hdr) -> dict:
    default_values = {"str": "default", "int": -999, "float": -999.9, "bool": False}
    key_is_copied = key_schema.get("copy")
    copy_schema_only = key_schema.get("copy") == "schema"
    key_is_renamed = key_schema.get("rename")
    renamed_key_is_in_header = key_is_renamed and (key_is_renamed in hdr)
    key_is_required = key_schema.get("required")
    key_is_in_header = key in hdr

    if copy_schema_only and key_is_in_header:
        return {key: default_values[key_schema["type"]]}
    if key_is_copied and key_is_in_header:
        return {key: hdr[key]}
    if key_is_copied and not key_is_in_header and renamed_key_is_in_header:
        return {key: hdr[key_schema.get("rename")]}
    if (
        key_is_copied
        and not key_is_in_header
        and key_is_renamed
        and not renamed_key_is_in_header
        and key_is_required
    ):
        raise KeyError(f" Required keyword {key!r} not found.")
    if key_is_copied and not key_is_in_header and not key_is_renamed and key_is_required:
        raise KeyError(f" Required keyword {key!r} not found.")
    if not key_is_copied and key_is_required and key_is_in_header:
        return {key: hdr[key]}
    if not key_is_copied and key_is_required and not key_is_in_header:
        return {key: default_values[key_schema["type"]]}
    # nothing to translate
    return {}



"""
Common Exceptions for the fits-validator
"""


__all__ = [
    "ValidationException",
    "SpecSchemaDefinitionException",
    "SpecValidationException",
    "ReturnTypeException",
]


class ValidationException(Exception):
    """
    Base Exception for the validator
    """

    def __init__(self, message: str = "Errors during validation", errors: dict = None):
        self.message = message
        self.errors = errors

    def __str__(self):
        return f"{self.message}: errors={self.errors}"


class SpecSchemaDefinitionException(ValidationException):
    """
    Exception when validating the YAML Schemas
    """


class SpecValidationException(ValidationException):
    """
    Default spec validation exception
    """


class ReturnTypeException(Exception):
    """
    Exception when returning a BytesIO object
    """

    def __init__(self, message: str = "Return type error"):
        self.message = message



from io import BytesIO
from pathlib import Path

import pytest
from astropy.io import fits

from dkist_header_validator import ReturnTypeException
from dkist_header_validator import spec122_validator
from dkist_header_validator import Spec122ValidationException


def test_translate_spec122_to_214_l0_base(valid_spec_122_object):
    """
    Validates and tries to translate a fits header against the SPEC-122 schema
    Given: A valid SPEC-122 fits header
    When: Validating headers
    Then: return translated HDUList and do not raise an exception
    """
    # raises exception on failure
    spec122_validator.validate_and_translate_to_214_l0(valid_spec_122_object)


def test_translate_required_only_headers_to_l0(valid_spec_122_object_required_only):
    """
    Validates and translates a spec122 compliant header with only the keywords required by the DC
    Given: A spec122 compliant header with only required keywords
    When: Validating and translating headers
    Then: return a translated 214l0 HDUList and do not raise an exception
    """
    spec122_validator.validate_and_translate_to_214_l0(valid_spec_122_object_required_only)


def test_translate_expected_only_headers_to_l0(valid_spec_122_object_expected_only):
    """
    Validates and translates a spec122 compliant header with only the keywords expected by the DC
    Given: A spec122 compliant header with only expected keywords
    When: Validating and translating headers
    Then: return a translated 214l0 HDUList and do not raise an exception
    """
    spec122_validator.validate_and_translate_to_214_l0(valid_spec_122_object_expected_only)


def test_translate_spec122_to_214_l0_return_dictionary(
    valid_spec_122_object,
):
    """
    Validates and tries to translate a fits header against the SPEC-122 schema
    Given: A valid SPEC-122 fits header
    When: Validating headers
    Then: return translated dictionary and do not raise an exception
    """
    # raises exception on failure
    result = spec122_validator.validate_and_translate_to_214_l0(
        valid_spec_122_object, return_type=dict
    )
    assert isinstance(result, dict)


def test_translate_spec122_to_214_l0_return_header(
    valid_spec_122_object,
):
    """
    Validates and tries to translate a fits header against the SPEC-122 schema
    Given: A valid SPEC-122 fits header
    When: Validating headers
    Then: return translated fits.header.Header object and do not raise an exception
    """
    # raises exception on failure
    result = spec122_validator.validate_and_translate_to_214_l0(
        valid_spec_122_object, return_type=fits.header.Header
    )
    assert isinstance(result, fits.header.Header)


def test_translate_spec122_to_214_l0_return_bytesio(valid_spec_122_file):
    """
    Validates and tries to translate a fits header against the SPEC-122 schema
    Given: A valid SPEC-122 fits header
    When: Validating headers
    Then: return translated BytesIO object and do not raise an exception
    """
    # raises exception on failure
    result = spec122_validator.validate_and_translate_to_214_l0(
        valid_spec_122_file, return_type=BytesIO
    )
    assert isinstance(result, BytesIO)


def test_translate_spec122_to_214_l0_return_hdu(valid_spec_122_file):
    """
    Validates and tries to translate a fits header against the SPEC-122 schema
    Given: A valid SPEC-122 fits header
    When: Validating headers
    Then: return translated PrimaryHDU object and do not raise an exception
    """
    # raises exception on failure
    result = spec122_validator.validate_and_translate_to_214_l0(
        valid_spec_122_file, return_type=fits.PrimaryHDU
    )
    assert isinstance(result, fits.PrimaryHDU)


def test_translate_spec122_to_214_l0_return_file(valid_spec_122_file):
    """
    Validates and tries to translate a fits file against the SPEC-122 schema
    Given: A valid SPEC-122 fits file
    When: Validating file
    Then: return translated file object and do not raise an exception
    """
    # raises exception on failure
    result = spec122_validator.validate_and_translate_to_214_l0(
        valid_spec_122_file, return_type=Path
    )
    assert isinstance(result, Path)


def test_spec122_to_214_l0_extrakeys_allowed(valid_spec_122_object_extra_keys):
    """
    Validates a fits header against the SPEC-0122 schema
    Given: A valid SPEC-0122 fits header with extra keys
    When: Validating headers
    Then: return HDUList and do not raise an exception
    """
    # raises exception on failure
    spec122_validator.validate_and_translate_to_214_l0(valid_spec_122_object_extra_keys)


def test_spec122_to_214_l0_valid_extrakeys_not_allowed(valid_spec_122_object_extra_keys):
    """
    Validates a fits header against the SPEC-0122 schema
    Given: A valid SPEC-0122 fits header with extra keys
    When: Validating headers
    Then: Raise a Spec122ValidationException
    """
    with pytest.raises(Spec122ValidationException):
        spec122_validator.validate_and_translate_to_214_l0(
            valid_spec_122_object_extra_keys, extra=False
        )


def test_translate_compressed_spec122_to_214_l0(valid_spec_122_compressed):
    """
    Validates and translates a compressed spec122 compliant file
    Given: A valid compressed SPEC-0122 file
    When: Validating headers
    Then: return valid HDUList and do not raise an exception
    """
    result = spec122_validator.validate_and_translate_to_214_l0(valid_spec_122_compressed)
    assert isinstance(result, fits.HDUList)


def test_visp_translate_to_214_l0(valid_spec_122_visp):
    """
    Validates a visp fits header against the SPEC-122 schema
    Given: A valid visp SPEC-122 fits header
    When: Validating headers
    Then: return validated HDUList and do not raise an exception
    """
    # raises exception on failure
    spec122_validator.validate_and_translate_to_214_l0(valid_spec_122_visp, return_type=dict)


def test_translate_to_214_l0_return_primaryhdu(valid_spec_122_file):
    """
    Validates a fits file against the SPEC-122 schema
    Given: A valid SPEC-122 fits file
    When: Validating and translating headers
    Then: return validated PrimaryHDU and do not raise an exception
    """
    # raises exception on failure
    spec122_validator.validate_and_translate_to_214_l0(
        valid_spec_122_file, return_type=fits.PrimaryHDU
    )


def test_translate_to_214_l0_return_primaryhdu_fail(valid_spec_122_no_file):
    """
    Validates and tries to translate a fits header against the SPEC-122 schema
    Given: A valid visp SPEC-122 fits header
    When: Validating headers
    Then: Raise a Spec122ValidationException
    """
    with pytest.raises(ReturnTypeException):
        spec122_validator.validate_and_translate_to_214_l0(
            valid_spec_122_no_file, return_type=fits.PrimaryHDU
        )


def test_translate_to_214_l0_datainsecondhdu(valid_spec_122_second_hdu):
    """
    Validates headers with data stored in second HDU
    Given: A valid SPEC-122 file or with data stored in second HDU
    When: Validating and translating headers
    Then: Raise an exception
    """
    # raises exception on failure
    spec122_validator.validate_and_translate_to_214_l0(valid_spec_122_second_hdu, return_type=Path)


def test_translate_spec122_to_214_l0_check_remaining_122_keys(valid_spec_122_dict_only):
    """
    Validates and tries to translate a fits header against the SPEC-122 schema.
    Checks to make sure that no original 122 keys were dropped.
    Given: A valid SPEC-122 fits header
    When: Validating headers
    Then: return translated HDUList and do not raise an exception
    """
    # raises exception on failure
    hdr = spec122_validator.validate_and_translate_to_214_l0(
        valid_spec_122_dict_only, return_type=dict
    )
    for key in valid_spec_122_dict_only.keys():
        if key not in hdr.keys():
            raise KeyError(f" Keyword {key!r} from original header dropped during translation!")


def test_translate_bytesio_spec122_to_214_l0_return_primaryhdu(valid_spec_122_hdulist_only):
    """
    Validates and tries to translate a BytesIO object to a 214 l0 PrimaryHDU.
    Given: A valid SPEC-122 BytesIO object
    When: Validating and translating headers
    Then: return translated PrimaryHDU and do not raise an exception
    """

    target = BytesIO()
    valid_spec_122_hdulist_only.writeto(
        target, output_verify="exception", overwrite=True, checksum=True
    )
    target.seek(0)
    spec122_validator.validate_and_translate_to_214_l0(
        input_headers=target, return_type=fits.PrimaryHDU, extra=True
    )



"""
Test for the base validator
"""
from pathlib import Path

import numpy as np
import pytest
import voluptuous as vol
import yaml
from astropy.io import fits
from deepdiff import DeepDiff

from dkist_header_validator.base_validator import SpecSchema
from dkist_header_validator.base_validator import SpecValidator
from dkist_header_validator.exceptions import SpecSchemaDefinitionException
from dkist_header_validator.exceptions import SpecValidationException
from dkist_header_validator.exceptions import ValidationException


@pytest.fixture(scope="module")
def valid_definition_format_params(tmpdir_factory):
    """
    Create a dict of valid schema definition formats to be used in
    successful format tests below.
    """

    dict1 = {
        "key_001": {
            "expand": True,
            "required": False,
            "comment": "comment 1",
            "type": "float",
            "values": [
                1.2,
                3.1,
            ],
        },
        "key_002": {
            "expand": False,
            "required": True,
            "comment": "comment 2",
            "type": "str",
        },
    }

    dict2 = {
        "key_003": {
            "expand": False,
            "required": True,
            "comment": "comment 3",
            "type": "int",
        },
        "key_004": {
            "expand": True,
            "required": False,
            "comment": "comment 4",
            "type": "bool",
        },
    }

    test_params = {
        "valid list[dicts]": [{**dict1, **dict2}],
    }

    yield test_params


@pytest.fixture(
    scope="function",
    params=[
        "valid list[dicts]",
    ],
)
def valid_definition_format(request, valid_definition_format_params):
    yield valid_definition_format_params[request.param]


def test_parse_spec_schema_definitions(valid_definition_format):
    """
    Given: A valid schema definition, can be in one of the following formats:
        list[dicts]
    When: Validating the schema
    Then: SpecSchema(definition) returns the schema definition in dict form
    """
    SpecSchema._parse_spec_schema_definitions(valid_definition_format)


@pytest.fixture(scope="module")
def invalid_definition_format_params(tmpdir_factory):
    """
    Create a dict of invalid schema definition formats to be used in
    failing format tests below.
    """

    invalid_list = [{"specxxx": {"not valid"}}, {"this is not a valid yaml file"}]

    temp_dir = tmpdir_factory.mktemp("invalid_definitions_temp")
    file_name = temp_dir.join("invalid_yaml_file.yml")
    file_name.write(yaml.safe_dump_all(invalid_list))

    test_params = {
        "invalid Dict": dict(),
        "invalid List[Dict]": [dict(), dict()],
        "invalid schema type1": "non valid schema type",
        "empty list": [],
    }

    yield test_params


@pytest.fixture(
    scope="function",
    params=[
        "invalid Dict",
        "invalid List[Dict]",
        "invalid schema type1",
        "empty list",
    ],
)
def invalid_definition_format(request, invalid_definition_format_params):
    yield invalid_definition_format_params[request.param]


def test_parse_spec_schema_definitions_fail(invalid_definition_format):
    """
    Given: A schema definition that is empty or not in one of the valid formats:
        List[dict]
    When: Validating the schema
    Then: raises a SpecSchemaDefinitionException
    """
    with pytest.raises(SpecSchemaDefinitionException):
        SpecSchema._parse_spec_schema_definitions(invalid_definition_format)


@pytest.fixture(scope="module")
def valid_schema_definition_params():
    """
    Create a dict of schema definition parameters that spans the valid
    parameter space. To be used in successful schema definition tests below.
    """

    test_params = {
        "required int": [{"keyword_name": {"required": True, "type": "int"}}],
        "optional int": [{"keyword_name": {"required": False, "type": "int"}}],
        "required float": [{"keyword_name": {"required": True, "type": "float"}}],
        "optional float": [{"keyword_name": {"required": False, "type": "float"}}],
        "required str": [{"keyword_name": {"required": True, "type": "str"}}],
        "optional str": [{"keyword_name": {"required": False, "type": "str"}}],
        "required bool": [{"keyword_name": {"required": True, "type": "bool"}}],
        "optional bool": [{"keyword_name": {"required": False, "type": "bool"}}],
        "optional key-value pair 1": [
            {"keyword_name": {"required": True, "type": "int", "optional1": "value1"}}
        ],
        "optional key-value pair 2": [
            {"keyword_name": {"required": False, "type": "int", "optional2": "value2"}}
        ],
        "optional key-value pair 3": [
            {"keyword_name": {"required": True, "type": "float", "optional3": "value3"}}
        ],
        "optional key-value pair 4": [
            {"keyword_name": {"required": False, "type": "float", "optional4": "value4"}}
        ],
        "optional key-value pair 5": [
            {"keyword_name": {"required": True, "type": "str", "optional5": "value5"}}
        ],
        "optional key-value pair 6": [
            {"keyword_name": {"required": False, "type": "str", "optional6": "value6"}}
        ],
        "optional key-value pair 7": [
            {"keyword_name": {"required": True, "type": "bool", "optional7": "value7"}}
        ],
        "optional key-value pair 8": [
            {"keyword_name": {"required": False, "type": "bool", "optional8": "value8"}}
        ],
    }

    yield test_params


@pytest.fixture(
    scope="function",
    params=[
        "required int",
        "optional int",
        "required float",
        "optional float",
        "required str",
        "optional str",
        "required bool",
        "optional bool",
        "optional key-value pair 1",
        "optional key-value pair 2",
        "optional key-value pair 3",
        "optional key-value pair 4",
        "optional key-value pair 5",
        "optional key-value pair 6",
        "optional key-value pair 7",
        "optional key-value pair 8",
    ],
)
def valid_schema_definition(request, valid_schema_definition_params):

    yield valid_schema_definition_params[request.param]


def test_validate_spec_schema_definitions(valid_schema_definition):
    """
    Given: Valid schema definitions that span the parameter space of the definition schema
    When: Validating the schema
    Then: SpecSchema(definition) successfully validates the definition against
        the definition schema. Success means no exception is raised.
    """
    SpecSchema(valid_schema_definition)


@pytest.fixture(scope="module")
def invalid_schema_definition_params():
    """
    Create a dict of invalid schema definition parameters to be used in
    failing tests below.
    """
    test_params = {
        "missing required": [
            {"specxxx": {"section": "spec_section", "copy": False}},
            {"keyword": {"type": "int"}},
        ],
        "missing type": [
            {"specxxx": {"section": "spec_section", "copy": False}},
            {"keyword": {"required": True}},
        ],
        "missing both, no optional": [
            {"specxxx": {"section": "spec_section", "copy": False}},
            {"keyword": {}},
        ],
        "missing both, with optional": [
            {"specxxx": {"section": "spec_section", "copy": False}},
            {"keyword": {"optional": "value"}},
        ],
    }

    yield test_params


@pytest.fixture(
    scope="function",
    params=[
        "missing required",
        "missing type",
        "missing both, no optional",
        "missing both, with optional",
    ],
)
def invalid_schema_definition(request, invalid_schema_definition_params):

    yield invalid_schema_definition_params[request.param]


def test_validate_spec_schema_definitions_fail(invalid_schema_definition):
    """
    Given: Schema definitions that do not meet the definition schema
    When: Validating the schema
    Then: SpecSchema(definition) raises a SpecSchemaDefinitionException
    """

    with pytest.raises(SpecSchemaDefinitionException):
        SpecSchema(invalid_schema_definition)


@pytest.fixture(scope="module")
def valid_values_list_params():
    """
    Create a dict of valid values lists to be used in successful tests below
    """
    test_params = {
        "str list": ["a", "b", "c"],
        "int list": [1, 2, 3],
        "float list": [1.0, 2.0, 3.0],
        "bool list": [True, False, False, True],
        "mixed list": ["a", 1, 2.0, True],
    }

    yield test_params


@pytest.fixture(
    scope="function", params=["str list", "int list", "float list", "bool list", "mixed list"]
)
def valid_values_list(request, valid_values_list_params):
    yield valid_values_list_params[request.param]


def test_validate_spec_schema_definitions_values_list(valid_values_list):
    """
    Given: Schema definitions that have valid values lists
    When: Validating the schema
    Then: SpecSchema(definition) does not raise a SpecSchemaDefinitionException
    """
    test_schema = [{"keyword_name": {"required": True, "type": "int", "values": valid_values_list}}]
    SpecSchema(test_schema)


@pytest.fixture(scope="module")
def invalid_values_list_params():
    """
    Create a dict of invalid values 'lists' to be used in failing tests below
    """
    test_params = {
        "non-list (set)": {"a", 1, 2.0, True},
        "non-list (tuple)": ("a", 1, 2.0, True),
        "non-list (scalar)": 1,
        "non-list (string)": "some string",
    }

    yield test_params


@pytest.fixture(
    scope="function",
    params=["non-list (set)", "non-list (tuple)", "non-list (scalar)", "non-list (string)"],
)
def invalid_values_list(request, invalid_values_list_params):
    yield invalid_values_list_params[request.param]


def test_validate_spec_schema_definitions_values_list_fail(invalid_values_list):
    """
    Given: Schema definitions that have invalid values lists
    When: Validating the schema
    Then: SpecSchema(definition) raises a SpecSchemaDefinitionException
    """
    test_schema = [
        {"specxxx": {"section": "spec_section", "copy": False}},
        {"keyword_name": {"required": True, "type": "int", "values": invalid_values_list}},
    ]
    with pytest.raises(SpecSchemaDefinitionException):
        SpecSchema(test_schema)


@pytest.fixture(scope="module")
def define_test_schema_definition():
    """
    Create a test schema to be used tested against in both successful and failing
    tests below.
    """
    test_schema_definition = [
        {
            "key_001": {
                "expand": False,
                "required": False,
                "comment": "comment 1",
                "type": "float",
            },
            "key_002": {
                "expand": False,
                "required": True,
                "comment": "comment 2",
                "type": "str",
            },
            "key_003": {
                "expand": False,
                "required": True,
                "comment": "comment 3",
                "type": "int",
            },
            "key_004": {
                "expand": False,
                "required": False,
                "comment": "comment 4",
                "type": "bool",
            },
        }
    ]

    yield test_schema_definition


def test_create_spec_schema(define_test_schema_definition):
    """
    Given: A valid schema definition and a valid test schema for the definition
    When: Creating the schema
    Then: The valid test schema is successfully parsed by the schema definition
    """
    schema = SpecSchema(define_test_schema_definition)

    input_dict = {"key_002": "value_002", "key_003": 314159, "key_001": 3.14159}

    assert not DeepDiff(input_dict, schema(input_dict, extra=True), ignore_order=True)


def test_create_spec_schema_fail(define_test_schema_definition):
    """
    Given: A valid schema definition and an invalid test schema for the definition
    When: Creating the schema
    Then: The invalid test schema is not successfully parsed by the schema definition
        and a voluptuous exception is raised.
    """
    schema = SpecSchema(define_test_schema_definition)

    # Required key key_003 is missing:
    input_dict = [{"key_002": "value_002", "key_004": False, "key_001": 3.14159}]
    with pytest.raises(vol.MultipleInvalid):
        schema(input_dict, extra=True)


@pytest.fixture(scope="module")
def spec_validator():
    """
    Create a schema validator to test fits headers against below.
    """
    test_schema = [
        {
            "BITPIX": {
                "expand": False,
                "required": True,
                "type": "int",
                "values": [8, 16, 32, 64, -32, -64],
            },
            "NAXIS": {"expand": True, "required": True, "type": "int", "values": [3]},
            "NAXIS1": {"expand": False, "required": True, "type": "int"},
            "NAXIS2": {"expand": False, "required": True, "type": "int"},
            "NAXIS3": {"expand": False, "required": True, "type": "int", "default_value": 1},
        }
    ]

    spec_schema = SpecSchema(test_schema)
    spec_validator = SpecValidator(spec_schema)

    yield spec_validator


@pytest.fixture(scope="module")
def valid_test_header_params(tmpdir_factory):
    """
    Create a dict with a valid header in various formats to be successfully tested
    against the schema defined above
    """
    temp_dir = tmpdir_factory.mktemp("valid test_headers_temp")
    file_name = temp_dir.join("tmp_fits_file.fits")
    valid_array = np.ones((1, 1, 1), dtype=float)
    valid_hdu = fits.PrimaryHDU(valid_array)
    valid_hdu_list = fits.HDUList([valid_hdu])
    valid_hdu_list.writeto(str(file_name))

    yield {
        "valid_hdu_list": valid_hdu_list,
        "valid_fits_header": valid_hdu_list[0].header,
        "valid_fits_file": Path(file_name),
    }


@pytest.fixture(
    scope="function",
    params=["valid_hdu_list", "valid_fits_header", "valid_fits_file"],
)
def valid_test_header(request, valid_test_header_params):
    yield valid_test_header_params[request.param]


def test_spec_validator(spec_validator, valid_test_header):
    """
    Given: A valid schema and a valid fits header to test against the schema
    When: Header validation
    Then: SpecValidator successfully validates the header and does not raise an exception
    """
    spec_validator.validate(valid_test_header)


@pytest.fixture(scope="module")
def valid_test_compressed_file(tmpdir_factory):
    """
    Create a compressed file with a valid header in various formats to be successfully tested
    against the schema defined above
    """
    temp_dir = tmpdir_factory.mktemp("valid test_headers_temp")
    file_name = temp_dir.join("tmp_fits_file.fits")
    valid_array = np.ones((1, 1, 1), dtype=float)
    primary_hdu = fits.PrimaryHDU()
    valid_hdu = fits.CompImageHDU(valid_array)
    valid_hdu_list = fits.HDUList([primary_hdu, valid_hdu])
    valid_hdu_list.writeto(str(file_name))

    yield {
        "valid_compressed.fits.fz": Path(file_name),
    }


@pytest.fixture(
    scope="function",
    params=["valid_compressed.fits.fz"],
)
def valid_test_compressed_files(request, valid_test_compressed_file):
    yield valid_test_compressed_file[request.param]


def test_spec_validator_file(spec_validator, valid_test_compressed_files):
    """
    Given: A valid schema and a valid compressed fits file to test against the schema
    When: Header validation
    Then: SpecValidator successfully validates the header and does not raise an exception
    """
    spec_validator.validate(valid_test_compressed_files)


@pytest.fixture(scope="module")
def invalid_test_header_params(tmpdir_factory):
    """
    Create a dict with an invalid header to be tested against the schema defined above.
    The header is invalid because the schema requires NAXIS to be 3 and here we create
    a 2D array which means NAXIS=2.
    """
    temp_dir = tmpdir_factory.mktemp("invalid_test_headers_temp")
    file_name = temp_dir.join("tmp_fits_file.fits")
    invalid_array = np.ones((1, 1), dtype=float)
    primary_hdu = fits.PrimaryHDU()
    invalid_hdu = fits.CompImageHDU(invalid_array)
    invalid_hdu_list = fits.HDUList([primary_hdu, invalid_hdu])
    invalid_hdu_list.writeto(str(file_name))

    yield {
        "invalid_hdu_list": invalid_hdu_list,
        "invalid_fits_header": invalid_hdu_list[0].header,
        "invalid_fits_file": Path(file_name),
    }


@pytest.fixture(
    scope="function", params=["invalid_hdu_list", "invalid_fits_header", "invalid_fits_file"]
)
def invalid_test_header(request, invalid_test_header_params):
    yield invalid_test_header_params[request.param]


def test_spec_validator_fail(spec_validator, invalid_test_header):
    """
    Given: A valid schema and an invalid fits header to test against the schema
    When: Header validation
    Then: SpecValidator raises a SpecValidationException
    """
    with pytest.raises(SpecValidationException):
        spec_validator.validate(invalid_test_header)


def test_validation_exception():
    """
    Test ValidationException string format
    Given: a ValidationException class
    When: instance created with message and errors
    Then: String of ValidationException is as expected
    """
    message = "test message"
    errors = {"error1": "error text"}
    s = ValidationException(message, errors=errors)
    assert str(s) == f"{message}: errors={errors}"



import pytest

from dkist_header_validator.translator import sanitize_to_spec214_level1
from dkist_header_validator.translator import translate_spec122_to_spec214_l0


def test_spec122_to_214_l0_valid(valid_translator_object):
    """
    Strips headers down to a 214 L1 object
    Given: A valid SPEC-0122 object
    When: Sanitizing headers
    Then:
    """
    translate_spec122_to_spec214_l0(valid_translator_object)


def test_spec122_to_214_l0_missing_required_keys(invalid_spec_122_object):
    """
    Translates an invalid SPEC-0122 object missing required
    keys to a SPEC-214 l0 object
    Given: A valid SPEC-0122 object
    When: Translating headers
    Then: Raises a KeyError exception
    """
    with pytest.raises(KeyError):
        translate_spec122_to_spec214_l0(invalid_spec_122_object)


def test_translate_to_214_l0_required_only_headers(valid_translator_object_required_only):
    """
    Translates a spec122 compliant header with only the keywords required by the DC
    Given: A spec122 compliant header with only required keywords
    When: Translating headers
    Then: For a fits file input, return a HDUList and do not raise an exception
          For a dict, HDUList, or header input, return a dictionary and do not raise an exception
    """
    translate_spec122_to_spec214_l0(valid_translator_object_required_only)


def test_translate_to_214_l0_expected_only_headers(valid_translator_object_expected_only):
    """
    Translates a spec122 compliant header with only the keywords required by the DC
    Given: A spec122 compliant header with only required keywords
    When: Translating headers
    Then: For a fits file input, return a HDUList and do not raise an exception
          For a dict, HDUList, or header input, return a dictionary and do not raise an exception
    """
    translate_spec122_to_spec214_l0(valid_translator_object_expected_only)


# I removed all of the compressed tests because a file (compressed or not) cannot go directly into the translator.
# It has to go into the validator first, which will hand it to the translator in the right format


def test_sanitize_s122(valid_translator_object):
    """
    Tries to sanitize a Spec-0122 object
    Given: A valid SPEC-0122 object
    When: Stripping down headers to 214 L1 (no 122 headers)
    Then: Raises a TypeError exception
    """
    with pytest.raises(TypeError):
        sanitize_to_spec214_level1(valid_translator_object)


def test_sanitize_translated_s214l0(valid_translator_object):
    """
    Tries to sanitize a Spec-214 L0 object
    Given: A valid SPEC-0122 object
    When: Translating and stripping down headers to 214 L1 (no 122 headers)
    Then: Raises a TypeError exception
    """
    s214_l0_header = translate_spec122_to_spec214_l0(valid_translator_object)
    with pytest.raises(TypeError):
        sanitize_to_spec214_level1(s214_l0_header)


def test_sanitize(valid_spec_214_no_file):
    """
    Sanitizes a SPEC-214 L1 object
    Given: A valid SPEC-214 L1 object
    When: Stripping down headers to 214 L1 (no 122 headers)
    Then: For a header or a dict, return a fits header and do not raise an exception
          For an HDUList or a file, return an HDUList and do not raise an exception
    """
    sanitize_to_spec214_level1(valid_spec_214_no_file)



"""
Validators configured for specific Fits Specs
"""
from dkist_fits_specifications import spec122
from dkist_fits_specifications import spec214
from dkist_fits_specifications.spec214 import level0

from dkist_header_validator.base_validator import SpecValidator
from dkist_header_validator.exceptions import SpecSchemaDefinitionException
from dkist_header_validator.exceptions import SpecValidationException

__all__ = [
    "spec122_validator",
    "Spec122ValidationException",
    "spec214_validator",
    "spec214_l0_validator",
    "Spec214ValidationException",
]


############
# SPEC-122 #
############


class Spec122ValidationException(SpecValidationException):
    """
    Exception when validating a spec 122 file
    """


spec122 = list(spec122.load_spec122().values())

spec122_validator = SpecValidator(
    spec_schema=spec122,
    SchemaValidationException=Spec122ValidationException,
)


############
# SPEC-214 #
############
class Spec214ValidationException(SpecValidationException):
    """
    Exception when validating a spec 214 file
    """


spec214_l0 = list(level0.load_level0_spec214().values())

spec214_l0_validator = SpecValidator(
    spec_schema=spec214_l0,
    SchemaValidationException=Spec214ValidationException,
)

spec214 = list(spec214.load_spec214().values())


spec214_validator = SpecValidator(
    spec_schema=spec214,
    SchemaValidationException=Spec214ValidationException,
)



from io import BytesIO
from pathlib import Path

from astropy.io import fits

from dkist_header_validator import spec122_validator


def test_spec122(valid_spec_122_object):
    """
    Validates a fits header against the SPEC-122 schema
    Given: A valid SPEC-122 fits header
    When: Validating headers
    Then: return validated HDUList and do not raise an exception
    """
    # raises exception on failure
    spec122_validator.validate(valid_spec_122_object, extra=False)


def test_spec122_return_dictionary(valid_spec_122_object):
    """
    Validates a fits header against the SPEC-122 schema
    Given: A valid SPEC-122 fits header
    When: Validating headers
    Then: return validated dictionary and do not raise an exception
    """
    # raises exception on failure
    result = spec122_validator.validate(valid_spec_122_object, return_type=dict, extra=False)
    assert isinstance(result, dict)


def test_spec122_return_fits_header(valid_spec_122_object):
    """
    Validates a fits header against the SPEC-122 schema
    Given: A valid SPEC-122 fits header
    When: Validating headers
    Then: return validated fits.header.Header object and do not raise an exception
    """
    # raises exception on failure
    result = spec122_validator.validate(
        valid_spec_122_object, return_type=fits.header.Header, extra=False
    )
    assert isinstance(result, fits.header.Header)


def test_spec122_return_BytesIO(valid_spec_122_file):
    """
    Validates a fits header against the SPEC-122 schema
    Given: A valid SPEC-122 fits header
    When: Validating headers
    Then: return validated BytesIO object and do not raise an exception
    """
    # raises exception on failure
    result = spec122_validator.validate(valid_spec_122_file, return_type=BytesIO, extra=False)
    assert isinstance(result, BytesIO)


def test_spec122_return_HDU(valid_spec_122_file):
    """
    Validates a fits header against the SPEC-122 schema
    Given: A valid SPEC-122 fits header
    When: Validating headers
    Then: return validated PrimaryHDU object and do not raise an exception
    """
    # raises exception on failure
    result = spec122_validator.validate(
        valid_spec_122_file, return_type=fits.PrimaryHDU, extra=False
    )
    assert isinstance(result, fits.PrimaryHDU)


def test_spec122_return_file(valid_spec_122_file):
    """
    Validates a fits header against the SPEC-122 schema
    Given: A valid SPEC-122 fits header
    When: Validating headers
    Then: return validated file and do not raise an exception
    """
    # raises exception on failure
    result = spec122_validator.validate(valid_spec_122_file, return_type=Path, extra=False)
    assert isinstance(result, Path)


def test_validate_required_only_headers(valid_spec_122_object_required_only):
    """
    Validates a spec122 compliant header with only the keywords required by the DC
    Given: A spec122 compliant header with only required keywords
    When: Validating headers
    Then: return a validated HDUList and do not raise an exception
    """
    spec122_validator.validate(valid_spec_122_object_required_only)


def test_validate_expected_only_headers(valid_spec_122_object_expected_only):
    """
    Validates a spec122 compliant header with only the keywords expected by the DC
    Given: A spec122 compliant header with only exxpected keywords
    When: Validating headers
    Then: return a validated HDUList and do not raise an exception
    """
    spec122_validator.validate(valid_spec_122_object_expected_only)


def test_compressed_spec122_valid(valid_spec_122_compressed):
    """
    Validates a compressed spec122 compliant file
    Given: A valid compressed SPEC-0122 file
    When: Validating headers
    Then: return valid HDUList and do not raise an exception
    """
    spec122_validator.validate(valid_spec_122_compressed)


def test_visp_spec122(valid_spec_122_visp):
    """
    Validates a visp fits header against the SPEC-122 schema
    Given: A valid visp SPEC-122 fits header
    When: Validating headers
    Then: return validated HDUList and do not raise an exception
    """
    # raises exception on failure
    spec122_validator.validate(valid_spec_122_visp, extra=False)


def test_datainsecondHDU(valid_spec_122_second_hdu):
    """
    Validates headers with data stored in second HDU
    Given: A valid SPEC-122 file or with data stored in second HDU
    When: Validating headers
    Then: Do not raise an exception
    """
    # raises exception on failure
    spec122_validator.validate(valid_spec_122_second_hdu, return_type=Path)


def test_casesensitivity(valid_spec_122_casesensitive):
    """
    Validates headers with a keyword that will need a case sensitive change
    Given: Headers with a keyword that will need a case sensitive change
    When: Validating headers
    Then: Do not raise and exception
    """
    # raises exception on failure
    spec122_validator.validate(valid_spec_122_casesensitive)



from io import BytesIO
from pathlib import Path

import numpy as np
import pytest

from dkist_header_validator import spec214_l0_validator
from dkist_header_validator import spec214_validator
from dkist_header_validator import Spec214ValidationException
from dkist_header_validator.exceptions import ReturnTypeException
from dkist_header_validator.exceptions import ValidationException


def test_spec214l0_return_BytesIO_without_data(valid_spec_214l0_no_file):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: Raise return type exception
    """
    # raises exception on failure
    with pytest.raises(ReturnTypeException):
        result = spec214_l0_validator.validate(
            valid_spec_214l0_no_file, return_type=BytesIO, extra=False
        )
        assert isinstance(result, BytesIO)


def test_spec214_return_BytesIO_without_data(valid_spec_214_no_file):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: Raise return type exception
    """
    # raises exception on failure
    with pytest.raises(ReturnTypeException):
        result = spec214_validator.validate(
            valid_spec_214_no_file, return_type=BytesIO, extra=False
        )
        assert isinstance(result, BytesIO)


def test_spec214l0_return_file_without_data(valid_spec_214l0_no_file):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: raise a return type exception
    """
    # raises exception on failure
    with pytest.raises(ReturnTypeException):
        result = spec214_l0_validator.validate(
            valid_spec_214l0_no_file, return_type=Path, extra=False
        )
        assert isinstance(result, Path)


def test_spec214_return_file_without_data(valid_spec_214_no_file):
    """
    Validates a fits header against the SPEC-0214 schema
    Given: A valid SPEC-0214 fits header
    When: Validating headers
    Then: raise a return type exception
    """
    # raises exception on failure
    with pytest.raises(ReturnTypeException):
        result = spec214_validator.validate(valid_spec_214_no_file, return_type=Path, extra=False)
        assert isinstance(result, Path)


def test_spec214l0_invalid_headers(invalid_spec_214l0_object):
    """
    Validates an invalid fits header against the SPEC-214 schema
    Given: A invalid SPEC-214 fits header
    When: Validating headers
    Then: raise a Spec214ValidationException
    """

    with pytest.raises(Spec214ValidationException):
        spec214_l0_validator.validate(invalid_spec_214l0_object)


def test_spec214_invalid_headers(invalid_spec_214_object):
    """
    Validates an invalid fits header against the SPEC-214 schema
    Given: A invalid SPEC-214 fits header
    When: Validating headers
    Then: raise a Spec214ValidationException
    """

    with pytest.raises(Spec214ValidationException):
        spec214_validator.validate(invalid_spec_214_object)


@pytest.fixture(scope="module")
def invalid_file_params(tmpdir_factory):
    """
    Create a dict of invalid file params to be used in failing
    tests below.
    """
    temp_dir = tmpdir_factory.mktemp("invalid_file_params_temp")
    non_existent_file_name = temp_dir.join("tmp_fits_file.fits")
    non_fits_file_name = temp_dir.join("tmp_this_is_not_a_fits_file.dat")
    temp_array = np.ones(1, dtype=np.int16)
    temp_array.tofile(str(non_fits_file_name))
    yield {"file not found": non_existent_file_name, "file_not_fits": non_fits_file_name}


@pytest.fixture(scope="function", params=["file not found", "file_not_fits"])
def invalid_file_param(request, invalid_file_params):
    yield invalid_file_params[request.param]


def test_file_errors(invalid_file_param):
    """
    Validates an invalid file spec
    Given: A invalid file specification: non-existent file or non-fits file
    When: Validating headers
    Then: raise a Spec214ValidationException
    """

    with pytest.raises(ValidationException):
        spec214_l0_validator.validate(invalid_file_param)


def test_invalid_compressed_spec214l0(invalid_spec_214l0_compressed):
    """
    Validates an invalid compressed spec214 compliant file
    Given: An invalid compressed SPEC-0214 file
    When: Validating headers
    Then: Catch a warning and raise an exception
    """
    with pytest.raises(Spec214ValidationException):
        spec214_l0_validator.validate(invalid_spec_214l0_compressed)


def test_invalid_compressed_spec214(invalid_spec_214_compressed):
    """
    Validates an invalid compressed spec214 compliant file
    Given: An invalid compressed SPEC-0214 file
    When: Validating headers
    Then: Catch a warning and raise an exception
    """
    with pytest.raises(Spec214ValidationException):
        spec214_validator.validate(invalid_spec_214_compressed)


def test_validatel0_toomanyHDUs(valid_spec_214_l0_too_many_HDUs):
    """
    Validates headers with too many (more than 2) HDUs
    Given: A valid SPEC-214 file or HDUList with more than two headers
    When: Validating headers
    Then: Raise an exception
    """
    # raises exception on failure
    with pytest.raises(ValidationException):
        spec214_l0_validator.validate(valid_spec_214_l0_too_many_HDUs)


def test_validate_toomanyHDUs(valid_spec_214_too_many_HDUs):
    """
    Validates headers with too many (more than 2) HDUs
    Given: A valid SPEC-214 file or HDUList with more than two headers
    When: Validating headers
    Then: Raise an exception
    """
    # raises exception on failure
    with pytest.raises(ValidationException):
        spec214_validator.validate(valid_spec_214_too_many_HDUs)



from pathlib import Path
from typing import Union
from uuid import uuid4

import numpy as np
import pytest
from astropy.io import fits
from astropy.wcs import WCS
from dkist_data_simulator.spec122 import Spec122Dataset
from dkist_data_simulator.spec214 import Spec214Dataset

FITS_OBJECT_TYPES = ["fits", "dict", "hdulist", "header"]


def get_fits_object(
    object_type: str, tmpdir: Path, ds, required_only: bool = False, expected_only: bool = False
) -> Union[str, dict, fits.HDUList, fits.header.Header]:
    """
    Build up the required object type from the given simulated Dataset

    TYPES
    -----
    fits : a fits file on disk
    compressed : a compressed fits file on disk
    dict : headers formatted as a python dict
    hdulist : headers built into a PrimaryHDU presented in an HDUList
    header : headers formatted as a fits.header.Header
    second_hdu: a fits file on disk with the headers located in the second hdu
    """
    header_dict = [d.header(required_only=required_only, expected_only=expected_only) for d in ds][
        0
    ]
    data = np.ones(shape=ds.array_shape)
    hdu = fits.PrimaryHDU(data=data)
    for key, value in header_dict.items():
        hdu.header[key] = value

    if object_type == "fits":
        file_name = tmpdir / f"{uuid4().hex}.fits"
        hdu_list = fits.HDUList([hdu])
        hdu_list.writeto(str(file_name))
        return str(file_name)

    if object_type == "compressed":
        file_name = tmpdir / f"{uuid4().hex}.fits"
        hdu_list = fits.HDUList(
            [fits.PrimaryHDU(), fits.CompImageHDU(data=data, header=hdu.header)]
        )
        hdu_list.writeto(str(file_name))
        return str(file_name)

    if object_type == "dict":
        return header_dict

    if object_type == "hdulist":
        return fits.HDUList([hdu])

    if object_type == "header":
        return hdu.header

    if object_type == "second_hdu":
        file_name = tmpdir / f"{uuid4().hex}.fits"
        hdu_list = fits.HDUList([fits.PrimaryHDU(), fits.ImageHDU(data=data, header=hdu.header)])
        hdu_list.writeto(str(file_name))
        return str(file_name)

    if object_type == "third_hdu":
        file_name = tmpdir / f"{uuid4().hex}.fits"
        hdu_list = fits.HDUList(
            [
                fits.PrimaryHDU(),
                fits.ImageHDU(data=data, header=hdu.header),
                fits.ImageHDU(data=data, header=hdu.header),
            ]
        )
        hdu_list.writeto(str(file_name))
        return str(file_name)


class BaseSpec122Dataset(Spec122Dataset):
    def __init__(self, instrument="vbi"):
        self.array_shape = (1, 10, 10)
        super().__init__(
            dataset_shape=(2, 10, 10),
            array_shape=self.array_shape,
            time_delta=1,
            instrument=instrument,
        )


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_122_object(tmpdir, request):
    yield get_fits_object(object_type=request.param, tmpdir=tmpdir, ds=BaseSpec122Dataset())


@pytest.fixture(scope="function", params=["dict", "header"])
def valid_translator_object(tmpdir, request):
    yield get_fits_object(object_type=request.param, tmpdir=tmpdir, ds=BaseSpec122Dataset())


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_122_object_required_only(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec122Dataset(), required_only=True
    )


@pytest.fixture(scope="function", params=["dict", "header"])
def valid_translator_object_required_only(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec122Dataset(), required_only=True
    )


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_122_object_expected_only(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec122Dataset(), expected_only=True
    )


@pytest.fixture(scope="function", params=["dict", "header"])
def valid_translator_object_expected_only(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec122Dataset(), expected_only=True
    )


@pytest.fixture(scope="function")
def valid_spec_122_file(tmpdir):
    yield get_fits_object(object_type="fits", tmpdir=tmpdir, ds=BaseSpec122Dataset())


@pytest.fixture(scope="function")
def valid_spec_122_compressed(tmpdir):
    yield get_fits_object(object_type="compressed", tmpdir=tmpdir, ds=BaseSpec122Dataset())


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_122_visp(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec122Dataset(instrument="visp")
    )


@pytest.fixture(scope="function", params=["dict", "header", "hdulist"])
def valid_spec_122_no_file(tmpdir, request):
    yield get_fits_object(object_type=request.param, tmpdir=tmpdir, ds=BaseSpec122Dataset())


@pytest.fixture(scope="function")
def valid_spec_122_dict_only(tmpdir):
    yield get_fits_object(object_type="dict", tmpdir=tmpdir, ds=BaseSpec122Dataset())


@pytest.fixture(scope="function")
def valid_spec_122_hdulist_only(tmpdir):
    yield get_fits_object(object_type="hdulist", tmpdir=tmpdir, ds=BaseSpec122Dataset())


@pytest.fixture(scope="function")
def valid_spec_122_second_hdu(tmpdir):
    yield get_fits_object(object_type="second_hdu", tmpdir=tmpdir, ds=BaseSpec122Dataset())


@pytest.fixture(scope="function")
def valid_spec_122_too_many_HDUs(tmpdir):
    yield get_fits_object(object_type="third_hdu", tmpdir=tmpdir, ds=BaseSpec122Dataset())


class Spec122DatasetExtraKeys(BaseSpec122Dataset):
    def __init__(self):
        super().__init__()
        self.add_constant_key("XTRAKEY1", "ABCDEFG")
        self.add_constant_key("XTRAKEY2", "HIJKLMN")
        self.add_constant_key("XTRAKEY3", "OPQRSTU")
        self.add_constant_key("XTRAKEY4", "VWXYZAB")


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_122_object_extra_keys(tmpdir, request):
    yield get_fits_object(object_type=request.param, tmpdir=tmpdir, ds=Spec122DatasetExtraKeys())


class InvalidBaseSpec122Dataset(Spec122Dataset):
    def __init__(self, instrument="vbi"):
        self.array_shape = (1, 10, 10)
        super().__init__(
            dataset_shape=(2, 10, 10),
            array_shape=self.array_shape,
            time_delta=1,
            instrument=instrument,
        )
        self.add_remove_key("NAXIS1")
        self.add_remove_key("NXAIS2")
        self.add_remove_key("WAVELNTH")
        self.add_remove_key("DATE-OBS")
        self.add_remove_key("ID___002")
        self.add_remove_key("ID___003")
        self.add_remove_key("ID___012")
        self.add_remove_key("DKIST003")
        self.add_remove_key("DKIST004")
        self.add_remove_key("HISTORY")


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def invalid_spec_122_object(tmpdir, request):
    yield get_fits_object(object_type=request.param, tmpdir=tmpdir, ds=InvalidBaseSpec122Dataset())


class BaseSpec122DatasetCaseSensitive(Spec122Dataset):
    def __init__(self, instrument="vbi"):
        self.array_shape = (1, 10, 10)
        super().__init__(
            dataset_shape=(2, 10, 10),
            array_shape=self.array_shape,
            time_delta=1,
            instrument=instrument,
        )
        self.add_constant_key("DKIST004", "oBSErve")


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_122_casesensitive(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec122DatasetCaseSensitive()
    )


class BaseSpec214l0Dataset(Spec122Dataset):
    def __init__(self, instrument="vbi"):
        self.array_shape = (1, 10, 10)
        super().__init__(
            dataset_shape=(2, 10, 10),
            array_shape=self.array_shape,
            time_delta=1,
            instrument=instrument,
            file_schema="level0_spec214",
        )


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_214l0_object(tmpdir, request):
    yield get_fits_object(object_type=request.param, tmpdir=tmpdir, ds=BaseSpec214l0Dataset())


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_214l0_object_required_only(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec214l0Dataset(), required_only=True
    )


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_214l0_object_expected_only(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec214l0Dataset(), expected_only=True
    )


@pytest.fixture(scope="function")
def valid_spec_214l0_file(tmpdir):
    yield get_fits_object(object_type="fits", tmpdir=tmpdir, ds=BaseSpec214l0Dataset())


@pytest.fixture(scope="function")
def valid_spec_214l0_compressed(tmpdir):
    yield get_fits_object(object_type="compressed", tmpdir=tmpdir, ds=BaseSpec214l0Dataset())


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_214l0_visp(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec214l0Dataset(instrument="visp")
    )


@pytest.fixture(scope="function", params=["dict", "header", "hdulist"])
def valid_spec_214l0_no_file(tmpdir, request):
    yield get_fits_object(object_type=request.param, tmpdir=tmpdir, ds=BaseSpec214l0Dataset())


@pytest.fixture(scope="function")
def valid_spec_214l0_dict_only(tmpdir):
    yield get_fits_object(object_type="dict", tmpdir=tmpdir, ds=BaseSpec214l0Dataset())


@pytest.fixture(scope="function")
def valid_spec_214l0_hdulist_only(tmpdir):
    yield get_fits_object(object_type="hdulist", tmpdir=tmpdir, ds=BaseSpec214l0Dataset())


@pytest.fixture(scope="function")
def valid_spec_214l0_second_hdu(tmpdir):
    yield get_fits_object(object_type="second_hdu", tmpdir=tmpdir, ds=BaseSpec214l0Dataset())


@pytest.fixture(scope="function")
def valid_spec_214_l0_too_many_HDUs(tmpdir):
    yield get_fits_object(object_type="third_hdu", tmpdir=tmpdir, ds=BaseSpec214l0Dataset())


class Spec214l0DatasetExtraKeys(BaseSpec214l0Dataset):
    def __init__(self):
        super().__init__()
        self.add_constant_key("XTRAKEY1", "ABCDEFG")
        self.add_constant_key("XTRAKEY2", "HIJKLMN")
        self.add_constant_key("XTRAKEY3", "OPQRSTU")
        self.add_constant_key("XTRAKEY4", "VWXYZAB")


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_214l0_object_extra_keys(tmpdir, request):
    yield get_fits_object(object_type=request.param, tmpdir=tmpdir, ds=Spec214l0DatasetExtraKeys())


class InvalidBaseSpec214l0Dataset(Spec122Dataset):
    def __init__(self, instrument="vbi"):
        self.array_shape = (1, 10, 10)
        super().__init__(
            dataset_shape=(2, 10, 10),
            array_shape=self.array_shape,
            time_delta=1,
            instrument=instrument,
            file_schema="level0_spec214",
        )
        self.add_remove_key("NAXIS1")
        self.add_remove_key("NXAIS2")
        self.add_remove_key("LINEWAV")
        self.add_remove_key("DATE-OBS")
        self.add_remove_key("FILE_ID")
        self.add_remove_key("EXPER_ID")


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def invalid_spec_214l0_object(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=InvalidBaseSpec214l0Dataset()
    )


@pytest.fixture(scope="function")
def invalid_spec_214l0_compressed(tmpdir):
    yield get_fits_object(object_type="compressed", tmpdir=tmpdir, ds=InvalidBaseSpec214l0Dataset())


class BaseSpec214l0DatasetCaseSensitive(Spec122Dataset):
    def __init__(self, instrument="vbi"):
        self.array_shape = (1, 10, 10)
        super().__init__(
            dataset_shape=(2, 10, 10),
            array_shape=self.array_shape,
            time_delta=1,
            instrument=instrument,
            file_schema="level0_spec214",
        )
        self.add_remove_key("VBISYNCM")
        self.add_constant_key("VBISYNCM", "fIxED")


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_214l0_casesensitive(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec214l0DatasetCaseSensitive()
    )


class BaseSpec214Dataset(Spec214Dataset):
    def __init__(self, instrument="vbi"):
        self.array_shape = (10, 10)
        super().__init__(
            dataset_shape=(2, 10, 10),
            array_shape=self.array_shape,
            time_delta=1,
            instrument=instrument,
        )

    @property
    def fits_wcs(self):
        w = WCS(naxis=2)
        w.wcs.crpix = self.array_shape[1] / 2, self.array_shape[0] / 2
        w.wcs.crval = 0, 0
        w.wcs.cdelt = 1, 1
        w.wcs.cunit = "arcsec", "arcsec"
        w.wcs.ctype = "HPLN-TAN", "HPLT-TAN"
        w.wcs.pc = np.identity(self.array_ndim)
        return w


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_214_object(tmpdir, request):
    yield get_fits_object(object_type=request.param, tmpdir=tmpdir, ds=BaseSpec214Dataset())


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_214_object_required_only(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec214Dataset(), required_only=True
    )


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_214_object_expected_only(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec214Dataset(), expected_only=True
    )


@pytest.fixture(scope="function")
def valid_spec_214_file(tmpdir):
    yield get_fits_object(object_type="fits", tmpdir=tmpdir, ds=BaseSpec214Dataset())


@pytest.fixture(scope="function")
def valid_spec_214_compressed(tmpdir):
    yield get_fits_object(object_type="compressed", tmpdir=tmpdir, ds=BaseSpec214Dataset())


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_214_visp(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec214Dataset(instrument="visp")
    )


@pytest.fixture(scope="function", params=["dict", "header", "hdulist"])
def valid_spec_214_no_file(tmpdir, request):
    yield get_fits_object(object_type=request.param, tmpdir=tmpdir, ds=BaseSpec214Dataset())


@pytest.fixture(scope="function")
def valid_spec_214_dict_only(tmpdir):
    yield get_fits_object(object_type="dict", tmpdir=tmpdir, ds=BaseSpec214Dataset())


@pytest.fixture(scope="function")
def valid_spec_214_hdulist_only(tmpdir):
    yield get_fits_object(object_type="hdulist", tmpdir=tmpdir, ds=BaseSpec214Dataset())


@pytest.fixture(scope="function")
def valid_spec_214_second_hdu(tmpdir):
    yield get_fits_object(object_type="second_hdu", tmpdir=tmpdir, ds=BaseSpec214Dataset())


@pytest.fixture(scope="function")
def valid_spec_214_too_many_HDUs(tmpdir):
    yield get_fits_object(object_type="third_hdu", tmpdir=tmpdir, ds=BaseSpec214Dataset())


class Spec214DatasetExtraKeys(BaseSpec214Dataset):
    def __init__(self):
        super().__init__()
        self.add_constant_key("XTRAKEY1", "ABCDEFG")
        self.add_constant_key("XTRAKEY2", "HIJKLMN")
        self.add_constant_key("XTRAKEY3", "OPQRSTU")
        self.add_constant_key("XTRAKEY4", "VWXYZAB")


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_214_object_extra_keys(tmpdir, request):
    yield get_fits_object(object_type=request.param, tmpdir=tmpdir, ds=Spec214DatasetExtraKeys())


class InvalidBaseSpec214Dataset(Spec214Dataset):
    def __init__(self, instrument="vbi"):
        self.array_shape = (10, 10)
        super().__init__(
            dataset_shape=(2, 10, 10),
            array_shape=self.array_shape,
            time_delta=1,
            instrument=instrument,
        )
        self.add_remove_key("NAXIS1")
        self.add_remove_key("NXAIS2")
        self.add_remove_key("LINEWAV")
        self.add_remove_key("DATE-OBS")
        self.add_remove_key("FILE_ID")
        self.add_remove_key("EXPER_ID")

    @property
    def fits_wcs(self):
        w = WCS(naxis=2)
        w.wcs.crpix = self.array_shape[1] / 2, self.array_shape[0] / 2
        w.wcs.crval = 0, 0
        w.wcs.cdelt = 1, 1
        w.wcs.cunit = "arcsec", "arcsec"
        w.wcs.ctype = "HPLN-TAN", "HPLT-TAN"
        w.wcs.pc = np.identity(self.array_ndim)
        return w


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def invalid_spec_214_object(tmpdir, request):
    yield get_fits_object(object_type=request.param, tmpdir=tmpdir, ds=InvalidBaseSpec214Dataset())


@pytest.fixture(scope="function")
def invalid_spec_214_compressed(tmpdir):
    yield get_fits_object(object_type="compressed", tmpdir=tmpdir, ds=InvalidBaseSpec214Dataset())


class BaseSpec214DatasetCaseSensitive(Spec214Dataset):
    def __init__(self, instrument="vbi"):
        self.array_shape = (10, 10)
        super().__init__(
            dataset_shape=(2, 10, 10),
            array_shape=self.array_shape,
            time_delta=1,
            instrument=instrument,
        )
        self.add_remove_key("VBISYNCM")
        self.add_constant_key("VBISYNCM", "fIxED")

    @property
    def fits_wcs(self):
        w = WCS(naxis=2)
        w.wcs.crpix = self.array_shape[1] / 2, self.array_shape[0] / 2
        w.wcs.crval = 0, 0
        w.wcs.cdelt = 1, 1
        w.wcs.cunit = "arcsec", "arcsec"
        w.wcs.ctype = "HPLN-TAN", "HPLT-TAN"
        w.wcs.pc = np.identity(self.array_ndim)
        return w


@pytest.fixture(scope="function", params=FITS_OBJECT_TYPES)
def valid_spec_214_casesensitive(tmpdir, request):
    yield get_fits_object(
        object_type=request.param, tmpdir=tmpdir, ds=BaseSpec214DatasetCaseSensitive()
    )



"""
Definition of the base objects for the creation of a spec validator
"""
import logging
import os
import pathlib
from io import BytesIO
from numbers import Integral
from numbers import Real
from pathlib import Path
from pathlib import PurePath
from typing import IO
from typing import List
from typing import Tuple
from typing import Type
from typing import Union

import astropy.time as t
import astropy.units as u
import numpy as np
import voluptuous as vol
from astropy.io import fits
from astropy.io.fits.hdu.hdulist import HDUList
from dkist_fits_specifications import spec214
from dkist_fits_specifications.utils import expand_naxis
from voluptuous.error import Invalid

from dkist_header_validator.exceptions import ReturnTypeException
from dkist_header_validator.exceptions import SpecSchemaDefinitionException
from dkist_header_validator.exceptions import SpecValidationException
from dkist_header_validator.exceptions import ValidationException
from dkist_header_validator.translator import translate_spec122_to_spec214_l0


logger = logging.getLogger(__name__)

__all__ = ["SpecValidator", "SpecSchema"]


class FormatInvalid(Invalid):
    """
    An value which does not match the schema format.
    """


class FormatValidator:
    """
    Validate that the format is valid.
    """

    supported_formats = ["isot", "unit"]

    def __init__(self, format_value):
        if format_value not in self.supported_formats:
            raise SpecSchemaDefinitionException(f"{format_value} is an unknown format to validate.")
        self.format_value = format_value

    def __call__(self, value):
        if self.format_value == "unit":
            try:
                return u.Unit(value, format="fits")
            except Exception:
                raise FormatInvalid(f"{value} is not a valid FITS unit")

        if self.format_value == "isot":
            try:
                return t.Time(value, format="fits")
            except Exception:
                raise FormatInvalid(f"{value} is not a valid FITS time")

    def __repr__(self):
        return f"FormatValidator(format='{self.format_value}')"


class FITSFloatInvalid(Invalid):
    """
    Validates a header value is a valid FITS float.

    This means it's not NaN or ±Inf.
    """


class FITSFloatValidator:
    """
    Validate a FITS floating point value.

    This validator assumes the type has also been validated independently.
    """

    def __call__(self, value):
        if not np.isfinite(value):
            raise FITSFloatInvalid(
                f"{value} is not finite, floats in a FITS header must be finite."
            )

        return value


class SpecSchema:
    """
    Define a schema that is used to validate FITS headers for a spec based upon structured definitions
    in YAML or dicts
    """

    # Schemas defined for a spec have the following structure per key
    definition_schema_definition = {
        vol.Required("required"): vol.Any(True, False),
        vol.Required("type"): vol.Any("int", "float", "str", "bool"),
        "values": list,
        "values_range": list,
        "expected": vol.Any(True, False),
        "expand": vol.Any(True, False),
        "format": vol.Any("unit", "isot"),
    }
    # Voluptuous schema instance used to validate the definitions
    definition_schema = vol.Schema(definition_schema_definition, extra=vol.ALLOW_EXTRA)

    def __init__(self, spec_schema_definitions: Union[Path, dict, List[dict]]):
        """
        Constructor for the SpecSchema which builds a voluptuous schema from
        specification definition files in YAML or dicts
        :param spec_schema_definitions: Definition of the spec's schema in one of the following forms
            - Dict definition of the spec schema
            - List of Dict definitions of the spec schema
            - Path to a YAML file defining the spec schema
            - Path to a directory containing YAML files defining spec schema
        """
        # convert spec schema definitions to a list of dicts
        self.spec_schema_definitions = self._parse_spec_schema_definitions(spec_schema_definitions)
        # validate the dict definition of the spec schema
        self._validate_spec_schema_definitions()

    @classmethod
    def _parse_spec_schema_definitions(cls, spec_schema: List[dict]) -> dict:
        """
        Convert from a list of dicts multiple formats to a dict
        :param spec_schema: Definition(s) of the spec's schema
        :return: Dictionary capturing the spec schema definition
        """

        # Test for proper types and non-emptiness
        if spec_schema:
            if all([isinstance(item, dict) and item for item in spec_schema]):
                schema = {key: schema for item in spec_schema for (key, schema) in item.items()}
                return schema
        # empty or not a known type
        raise SpecSchemaDefinitionException(
            "Schema definition is empty or is not a supported type",
            errors={
                "received": type(spec_schema),
                "expected": List[dict],
            },
        )

    def _validate_spec_schema_definitions(self) -> None:
        """
        Validate the spec schema definitions against the class schema for
        spec schema definitions and raise a SpecSchemaDefinitionException
        on failure
        :return: None
        """

        for key, spec_schema in self.spec_schema_definitions.items():
            schema_errors = {}
            try:
                self.definition_schema(spec_schema)
            except vol.MultipleInvalid as e:
                schema_errors = {error.path[0]: error.msg for error in e.errors}
            if schema_errors:
                logger.debug(
                    f"Errors during schema definition validation. key={key} errors={schema_errors}"
                )
                raise SpecSchemaDefinitionException(
                    f"Errors during schema definition validation. key={key}",
                    errors=schema_errors,
                )

    def expand_schema(self, headers) -> dict:
        """
        Expand schema indices using information from fits headers
        :param headers: Fits file headers whose values are to be used to know how to expand spec keys
        :return: schema dictionary
        """
        # 214 Expansion
        if {"DAAXES", "DEAXES", "DNAXIS"}.issubset(headers.keys()):
            # This function includes the expansion for NAXIS
            return spec214.expand_214_schema(self.spec_schema_definitions, **headers)

        # 122 Only expands "n" with NAXIS
        return expand_naxis(headers["NAXIS"], self.spec_schema_definitions)

    @staticmethod
    def generate_schema_for_key(key_schema):
        """
        Generate voluptuous schema for a key
        :param key_schema: Spec schema (from yml) used to generate voluptuous schema
        :return: Voluptuous schema for a key
        """
        type_map = {"int": Integral, "float": Real, "str": str, "bool": bool}

        checks = []
        # Always check type
        checks.append(type_map[key_schema.get("type")])

        def case_insenstive_values(value):
            allowed_values = key_schema.get("values")
            if key_schema.get("type") == "str":
                value = value.lower()
                allowed_values = [v.lower() for v in allowed_values]

            return vol.Any(*allowed_values)(value)

        def expand_values_range(value):
            min_value = key_schema.get("values_range")[0]
            try:
                max_value = key_schema.get("values_range")[1]
            except IndexError:
                max_value = None

            return vol.Range(min=min_value, max=max_value)(value)

        if key_schema.get("values"):
            checks.append(case_insenstive_values)

        if key_schema.get("values_range"):
            checks.append(expand_values_range)

        if "format" in key_schema:
            checks.append(FormatValidator(key_schema["format"]))

        if key_schema.get("type") == "float":
            checks.append(FITSFloatValidator())

        return vol.All(*checks)

    def _add_keys_to_schema(self, schema: dict) -> dict:
        schema_keys = {}  # keys to be added to voluptuous schema
        required_keys = {
            vol.Required(k): self.generate_schema_for_key(v)
            for k, v in schema.items()
            if v.get("required")
        }
        schema_keys.update(required_keys)
        remaining_keys = {
            k: self.generate_schema_for_key(v) for k, v in schema.items() if not v.get("required")
        }
        schema_keys.update(remaining_keys)
        return schema_keys

    def _check_for_expansion(self, headers):
        for value in self.spec_schema_definitions.values():
            # check the whole schema to see if there is an expand = True keyword
            if value["expand"]:
                # if there is an 'expand' keyword set to true in the schema, expand the whole spec schema
                expanded_schema = self.expand_schema(headers)
                return expanded_schema
        return self.spec_schema_definitions

    def _create_spec_schema(self, headers, extra) -> vol.Schema:
        """
        A voluptuous.spec_validator object to validate headers against.
        Constructed from Spec keywords.
        :param headers: Fits file headers
        :return: Voluptuous schema
        """

        spec_schema = {}

        schema = self._check_for_expansion(headers)
        spec_schema.update(self._add_keys_to_schema(schema))
        if extra:
            return vol.Schema(spec_schema, extra=vol.ALLOW_EXTRA)
        return vol.Schema(spec_schema)

    def __call__(self, headers: dict, extra) -> vol.Schema:
        """
        Validate headers against the instance spec schema
        raising voluptuous errors on failure
        :param headers: header dict to validate
        :return: vol.Schema
        """
        """for value in self.spec_schema_definitions.values():
            if value["values_range"]:
                pass"""

        spec_schema = self._create_spec_schema(headers, extra)
        return spec_schema(headers)


class SpecValidator:
    """
    Validates FITS Headers against a schema
    """

    def __init__(
        self,
        spec_schema: Union[Path, PurePath, dict, List[dict], SpecSchema],
        SchemaValidationException: Type[SpecValidationException] = SpecValidationException,
    ):
        """
        Constructor for the SpecValidator
        :param spec_schema: Definition of the spec's schema in one of the following forms
            - SpecSchema instance
            - Dict definition of the spec schema
            - List of Dict definitions of the spec schema
            - Path to a YAML file defining the spec schema
            - Path to a directory containing YAML files defining spec schema
        :param SchemaValidationException: SpecValidationException or subclass of SpecValidationException
            to raise if spec_validator validation fails
        """
        # Callable for validating a dict against the defined spec_validator
        if isinstance(spec_schema, SpecSchema):
            self.spec_schema = spec_schema
        else:
            self.spec_schema = SpecSchema(spec_schema)

        # Exception raised when spec validation fails
        self.SchemaValidationException = SchemaValidationException

    @staticmethod
    def _headers_to_dict(headers: Union[HDUList, dict, fits.header.Header, str, IO]) -> dict:
        """
        Convert headers from multiple types to a dict
        :param headers: Headers to convert to a dict
        :return: Dict of the headers
        """
        if isinstance(headers, dict):
            return headers
        if isinstance(headers, fits.header.Header):
            return dict(headers)
        if isinstance(headers, HDUList):
            if len(headers) > 1:
                return dict(headers[1].header)
            return dict(headers[0].header)

    def verify_headers(self, headers, extra) -> dict:
        """
        Validates file headers against the instance spec_validator
        :param headers: file headers
        :param extra: switch for validation to allow extra keys in schema
        :return: dict of headers
        :raises: SchemaValidationException
        """

        validation_errors = {}
        try:
            self.spec_schema(headers, extra)
        except vol.MultipleInvalid as e:
            for error in e.errors:
                value_str = "Required keyword not present"
                if error.path[0] in headers:
                    value_str = (
                        f"Actual value: {headers[error.path[0]]!r} {type(headers[error.path[0]])}"
                    )
                validation_errors[error.path[0]] = f"{error.msg}. {value_str}"
        # Raise exception if we have errors
        if validation_errors:
            missing_list = []
            badtype_list = []
            other_list = []

            for key in validation_errors.keys():
                if "required key not provided" in validation_errors[key]:
                    missing_list.append(key)
                elif "expected" in validation_errors[key]:
                    badtype_list.append(key)
                else:
                    other_list.append(key)
            # Log the bad keys
            for sublist, message in zip(
                [missing_list, badtype_list, other_list],
                [
                    "The following keys are missing:",
                    "The following keys have bad types:",
                    "The following keys have other errors:",
                ],
            ):
                if len(sublist) > 0:
                    logger.debug(f"\n{message}")
                    for k in sorted(sublist):
                        logger.debug(f"{str(k):<10}:\t{validation_errors[k]}")

            raise self.SchemaValidationException(errors=validation_errors)
        logger.debug("Schema validation succeeded")

        return headers

    def _validate_headers(
        self, input_headers: Union[HDUList, dict, fits.header.Header], extra
    ) -> Tuple[dict, dict]:
        """
        Validates open input headers against the instance spec_schema
        :param input_headers: The input headers to validate in the following formats:
            - HDUList object
            - fits.header.Header object
            - Dictionary of header keys and values
        :param extra: switch for validation to allow extra keys in schema
        :return: dictionary of verified headers to be used later
        """
        if isinstance(input_headers, HDUList):
            if len(input_headers) > 2:
                raise ValidationException(
                    "Too many HDUs in your HDUList! May only have two HDUs at most."
                )
        headers = self._headers_to_dict(input_headers)
        fits_cards = self._capture_fits_cards(headers)
        verified_headers = self.verify_headers(headers, extra)
        return verified_headers, fits_cards

    def _validate_file(self, input_headers: Union[str, IO], extra) -> Tuple[dict, dict, np.ndarray]:
        """
        Validates files against the astropy and then instance spec_schema
        :param input_headers: The input headers to validate in the following formats:
            - string file path
            - File like object
        :param extra: switch for validation to allow extra keys in schema
        :return: dictionary of verified headers to be used later
        """
        try:
            with fits.open(input_headers) as hdul:
                if len(hdul) > 2:
                    raise ValidationException(
                        "Too many HDUs in your HDUList! May only have two HDUs at most."
                    )
                # verify fits headers with astropy verify library
                hdul.verify("exception")
                # normalize headers into a dict
                try:
                    hdus = self._headers_to_dict(hdul[1].header)
                    data = hdul[1].data
                except IndexError:  # non-compressed
                    hdus = self._headers_to_dict(hdul[0].header)
                    data = hdul[0].data
                fits_cards = self._capture_fits_cards(hdus)
                verified_headers = self.verify_headers(hdus, extra)
                return verified_headers, fits_cards, data
        except (ValueError, FileNotFoundError, OSError, IndexError) as exc:
            logger.debug(f"Cannot parse headers: detail = {exc}")
            raise ValidationException(f"Cannot parse headers", errors={type(exc): str(exc)})

    @staticmethod
    def _return_hdulist(validated_headers, fits_cards) -> HDUList:
        """
        Returns validated headers as an HDUList
        :param validated_headers: Already validated/translated headers to be written out into
          an HDUList
        :param fits_cards: Any special cards to be included in the HDUList
        :return: HDUList
        """
        if isinstance(validated_headers, HDUList):
            return validated_headers
        temp_array: np.ndarray = np.ones((1, 1, 1), dtype=np.int16)
        new_hdu = fits.PrimaryHDU(temp_array, header=fits.Header())
        for (key, value) in validated_headers.items():
            new_hdu.header[key] = value
        for key in fits_cards:
            new_hdu.header[key] = str(fits_cards[key])
        new_hdu_list = fits.HDUList([new_hdu])
        return new_hdu_list

    @staticmethod
    def _return_dictionary(validated_headers, fits_cards) -> dict:
        """
        Returns validated headers as a dictionary
        :param validated_headers: Already validated/translated headers to be written out into
          a dictionary
        :param fits_cards: Any special cards to be included in the dictionary
        :return: dictionary
        """
        for key in fits_cards:
            validated_headers[key] = str(fits_cards[key])
        return validated_headers

    @staticmethod
    def _return_BytesIO(validated_headers, input_headers, data, fits_cards) -> BytesIO:
        """
        Returns validated headers as a BytesIO object
        :param validated_headers: Already validated/translated headers to be written out into
          the BytesIO object
        :param input_headers: original filepath or BytesIO object
        :param data: original data
        :param fits_cards: Any special cards to be included in the BytesIO object
        :return: BytesIO object
        """
        new_hdu = fits.PrimaryHDU(data, header=fits.Header())
        for (key, value) in validated_headers.items():
            new_hdu.header[key] = value
        for key in fits_cards:
            new_hdu.header[key] = str(fits_cards[key])
        new_hdu_list = fits.HDUList([new_hdu])
        return BytesIO(
            new_hdu_list.writeto(
                str(os.path.basename(input_headers)),
                overwrite=True,
                output_verify="exception",
                checksum=True,
            )
        )

    @staticmethod
    def _return_file(validated_headers, input_headers, data, fits_cards) -> (str, IO):
        """
        Returns validated headers as a FITS file
        :param validated_headers: Already validated/translated headers to be written out
          into a FITS file
        :param input_headers: original filepath or BytesIO object
        :param data: original data
        :param fits_cards: Any special cards to be included in the FITS file
        :return: FITS file
        """
        new_hdu = fits.PrimaryHDU(data, header=fits.Header())
        for (key, value) in validated_headers.items():
            new_hdu.header[key] = value
        for key in fits_cards:
            new_hdu.header[key] = str(fits_cards[key])
        new_hdu_list = fits.HDUList([new_hdu])
        new_hdu_list.writeto(
            str(os.path.basename(input_headers)),
            overwrite=True,
            output_verify="exception",
            checksum=True,
        )
        return Path(os.path.basename(input_headers))

    @staticmethod
    def _return_HDU(validated_headers, data, fits_cards):
        """
        Returns validated headers as an HDU
        :param validated_headers: Already validated/translated headers to be written out
          into a FITS file
        :param data: original data
        :param fits_cards: Any special cards to be included in the FITS file
        :return: HDU
        """
        new_hdu = fits.PrimaryHDU(data, header=fits.Header())
        for (key, value) in validated_headers.items():
            new_hdu.header[key] = value
        for key in fits_cards:
            new_hdu.header[key] = str(fits_cards[key])
        return new_hdu

    def _format_output(
        self, return_type, validated_headers, input_headers=None, data=None, fits_cards=None
    ):
        fits_cards = fits_cards or {}
        if return_type == Path:
            if data is None:
                raise ReturnTypeException("No data. Cannot write file.")
            return self._return_file(validated_headers, input_headers, data, fits_cards)
        if return_type == BytesIO:
            if data is None:
                raise ReturnTypeException("No data. Cannot write BytesIO object.")
            return self._return_BytesIO(validated_headers, input_headers, data, fits_cards)
        if return_type == dict:
            return self._return_dictionary(validated_headers, fits_cards)
        if return_type == HDUList:
            return self._return_hdulist(validated_headers, fits_cards)
        if return_type == fits.header.Header:
            return self._return_HDU(validated_headers, data, fits_cards).header
        if return_type == fits.PrimaryHDU:
            if data is None:
                raise ReturnTypeException("No data. Cannot write PrimaryHDU.")
            return self._return_HDU(validated_headers, data, fits_cards)

    def _capture_fits_cards(self, validated_headers) -> dict:
        """
        Pull special fits cards out of validated_headers dict.
        This is necessary for astropy header formatting.

        :param validated_headers: validated headers
        :return: fits_cards: dictionary containing special fits header keys and values
        """
        fits_cards = {}
        if "HISTORY" in validated_headers:
            fits_cards["HISTORY"] = str(validated_headers["HISTORY"])
            validated_headers.pop("HISTORY")
        if "COMMENT" in validated_headers:
            fits_cards["COMMENT"] = str(validated_headers["COMMENT"])
            validated_headers.pop("COMMENT")
        return fits_cards

    def validate(self, input_headers, return_type=HDUList, extra=True):
        """
        Validates against the instance spec_schema
        :param input_headers: The headers to validate in the following formats:
            - string file path
            - File like object
            - HDUList object
            - fits.header.Header object
            - Dictionary of header keys and values
        :param return_type: determines return type. Default is HDUList. May be one of:
            - dict
            - BytesIO
            - fits.header.Header
            - Path (file)
            - HDUList
            - fits.PrimaryHDU
        :param extra: switch for validation to allow extra keys in schema. Default is true, which will
        allow extra keys. Ingest validation should allow extra keys.
        :return: formatted headers
        :raises: SpecValidationException or subclass
        """
        if isinstance(input_headers, (dict, fits.header.Header, HDUList)):
            validated_headers, fits_cards = self._validate_headers(input_headers, extra)
            return self._format_output(
                return_type, validated_headers, input_headers, None, fits_cards
            )
        validated_headers, fits_cards, data = self._validate_file(input_headers, extra)
        return self._format_output(return_type, validated_headers, input_headers, data, fits_cards)

    def validate_and_translate_to_214_l0(self, input_headers, return_type=HDUList, extra=True):
        """
        Validates against the instance spec_schema and then translates to the spec214_l0 schema
        :param input_headers: The headers to validate in the following formats:
            - string file path
            - File like object
            - HDUList object
            - fits.header.Header object
            - Dictionary of header keys and values
        :param return_type: determines return type. Default is HDUList. May be one of:
            - dict
            - BytesIO
            - fits.header.Header
            - Path (file)
            - HDUList
            - fits.PrimaryHDU
        :param extra: switch for validation to allow extra keys in schema. Default is true, which will
        allow extra keys. Ingest validation should allow extra keys.
        :return: formatted 214 l0 headers
        :raises: SpecValidationException or subclass
        """
        if isinstance(input_headers, (dict, fits.header.Header, HDUList)):
            validated_headers, fits_cards = self._validate_headers(input_headers, extra)
            translated_headers = translate_spec122_to_spec214_l0(validated_headers)
            return self._format_output(return_type, translated_headers, None, None, fits_cards)
        else:
            validated_headers, fits_cards, data = self._validate_file(input_headers, extra)
            translated_headers = translate_spec122_to_spec214_l0(validated_headers)
            return self._format_output(
                return_type, translated_headers, input_headers, data, fits_cards
            )
